# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst
#
# Astropy documentation build configuration file.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this file.
#
# All configuration values have a default. Some values are defined in
# the global Astropy configuration which is loaded here before anything else. 
# See astropy.sphinx.conf for which values are set there.

import datetime
#import sys
#import os

import sphinx_rtd_theme

#sys.path.insert(0, os.path.abspath("static"))

# -- General configuration ----------------------------------------------------
# If your documentation needs a minimal Sphinx version, state it here.

needs_sphinx = '1.3'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.mathjax',
    'sphinx.ext.extlinks',
#    'sphinx.ext.linkcode',
    'jupyter_sphinx',
    'sphinx.ext.doctest',
    'numpydoc',
]


intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
    'astropy': ('http://docs.astropy.org/en/stable/', None),
    'synphot': ('https://synphot.readthedocs.io/en/latest/', None),
    }

extlinks = {'python': ('https://docs.python.org/3/%s', None),
            'numpy': ('https://docs.scipy.org/doc/numpy/', None),
            'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
            'astropy': ('http://docs.astropy.org/en/stable/', None),
            'synphot': ('https://synphot.readthedocs.io/en/latest/', None),
            }


numpydoc_show_class_members = False
autosummary_generate = True
autoclass_content = "class"
autodoc_default_flags = ["members", "inherited-members"]
autodoc_docstring_signature = False

source_suffix = '.rst'
source_encoding = 'utf-8-sig'

master_doc = 'index'


# -- Project information ------------------------------------------------------

# This does not *have* to match the package name, but typically does
project = 'spextra'
author = 'Miguel Verdugo'
current_year = datetime.datetime.now().year
copyright = '2020-{:d}, {}'.format(current_year, author)


# import spextra
# The short X.Y version.
version = "0.1" #spextra.__version__.split('-', 1)[0]
release = version
# The full version, including alpha/beta/rc tags.

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all
# documents.

default_role = 'obj'

# The name of the Pygments (syntax highlighting) style to use.

pygments_style = 'default' #'sphinx'

# -- Options for HTML output ---------------------------------------------------

html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = [sphinx_gallery.glr_path_static()]


# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = ''

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = '{0} v{1}'.format(project, release)

# Output file base name for HTML help builder.
htmlhelp_basename = project + 'doc'

# Add local templates path to modify autosummary templates
#templates_path = ['_templates']

# Static files to copy after template files
html_static_path = ['_static']


# -- Options for LaTeX output --------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).

latex_documents = [('index', project + '.tex', project + u' Documentation',
                    author, 'manual')]


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).

man_pages = [('index', project.lower(), project + u' Documentation',
              [author], 1)]




# -*- coding: utf-8 -*-
"""
Database for speXtra
"""
import os
import inspect
import shutil
from posixpath import join as urljoin

import yaml

from .utils import Config,  download_file, dict_generator

# Configurations

__all__ = ["Database", "SpecLibrary", "SpectrumContainer", "ExtCurvesLibrary",
           "ExtCurveContainer", "FilterSystem", "FilterContainer", "DefaultData"]

# Tables are displayed with a jsviewer by default
# Table.show_in_browser.__defaults__ = (5000, True, 'default', {'use_local_files': True},
#                                              None, 'display compact', None, 'idx')


class DataContainer:
    """
    Just a general container for the data

    It reads a YAML file and creates attributes with the kyes of the dictionary.

    Additionally it contains methods for nicely displaying the contents on the screen.
    """
    def __init__(self, filename):
        self.filename = filename
        with open(self.filename) as f:
            self.meta = yaml.safe_load(f)
        for key in self.meta:
            setattr(self, key, self.meta[key])

    def dump(self):
        """
        (Try to) nicely dump the contents of the library

        Returns
        -------

        """
        print(yaml.dump(self.meta,
                        indent=4, sort_keys=False, default_flow_style=False))


class Database(DataContainer):
    """
    This class contains the database.


    It also acts as a lazy fetcher for remote data.
    When asked for local absolute path to a file or directory, Database
    checks if the file or directory exists locally and, if so, returns it.
    If it doesn't exist, it first determines where to get it from.
    It first downloads the file ``{remote_root}/redirects.json`` and checks
    it for a redirect from ``{relative_path}`` to a full URL. If no redirect
    exists, it uses ``{remote_root}/{relative_path}`` as the URL.
    It downloads then downloads the URL to ``{rootdir}/{relative_path}``.
    For directories, ``.tar.gz`` is appended to the
    ``{relative_path}`` before the above is done and then the
    directory is unpacked locally.
    Parameters
    ----------
    rootdir : str or callable
        The local root directory, or a callable that returns the local root
        directory given no parameters. (The result of the call is cached.)
        Using a callable allows one to customize the discovery of the root
        directory (e.g., from a config file), and to defer that discovery
        until it is needed.
    remote_root : str
        Root URL of the remote server.
    """

    def __init__(self, silent=False):

        conf = Config()
        self.database_url = conf.get_database_url()

        if not self.database_url.endswith('/'):
            self.database_url = self.database_url + '/'

        self.data_dir = conf.get_data_dir()
        self.ymlfile = "index.yml"
        self.path = self.abspath(self.ymlfile, silent=silent)

        super().__init__(filename=self.path)

        self.liblist,  self.relpathlist = self._makelists()

    def abspath(self, relpath, reload=False, silent=False):
        """
        Return absolute path to file or directory, ensuring that it exists.
        If it doesn't exist it will download it from the remote host

        Otherwise, just look for ``{relpath}``.
        """

        abspath = os.path.join(self.data_dir, relpath)
        if (os.path.exists(abspath) is False) or (reload is True):
            print("updating/loading '%s'" % relpath )
            url = urljoin(self.database_url, relpath)
            download_file(url, abspath, silent=silent)

        return abspath

    def remove_database(self):
        """
        Remove database and all files
        """
        try:
            shutil.rmtree(self.data_dir)
            print("database at %s removed" % self.data_dir)
        except FileNotFoundError:
            print("database at %s doesn't exist" % self.data_dir)
            raise

    def update(self):
        """
        Force an update of the cached database file
        """
        self.abspath(self.ymlfile, reload=True,  silent=False)

    def _makelists(self):
        """
        just make lists of the libraries and paths in the database"
        """
        a = list(dict_generator(self.meta))
        separator = '/'
        libs = [e[1:] for e in a]
        liblist = [separator.join(e) for e in libs]
        relpathlist = [separator.join(e) for e in a]

        return liblist, relpathlist

    def __repr__(self):
        description = "Database:"
        remote = "url: " + self.database_url
        local = "path: " + self.data_dir

        return ' %s \n %s \n %s ' % (description, remote, local)

    def __dir__(self):
        str = "Database of spextra"
        return str


class DefaultData:
    """
    small class just to define defaults spectra and filters
    """
    def __init__(self, **kwargs):

        database = Database()
        self.default_filters_file = database.abspath("default_filters.yml")
        self.default_spectra_file = database.abspath("default_spectra.yml")
        self.default_curves_file = database.abspath("default_curves.yml")

        self.filters = self.setdict(self.default_filters_file, **kwargs)
        self.spectra = self.setdict(self.default_spectra_file, **kwargs)
        self.extcurves = self.setdict(self.default_curves_file, **kwargs)

    @staticmethod
    def setdict(ymlfile, **kwargs):
        database = Database()
        path = database.abspath(ymlfile, **kwargs)

        with open(path) as filename:
            dictionary = yaml.safe_load(filename)
        return dictionary

    @staticmethod
    def update():
        DefaultData(silent=False, reload=True)


class Library(DataContainer):
    """
    Class that contains the information of a Library, either spectral
    of a filter system, extinction curves, etc
    """
    def __init__(self, library_name):
        self.name = library_name
        database = Database()

        if library_name not in database.liblist:
            raise ValueError("Library '%s' not in the database" % library_name)
        else:
            index = database.liblist.index(library_name)
            self.relpath = database.relpathlist[index]

        self.ymlfile = "index.yml"
        self.path = database.abspath(os.path.join(self.relpath, self.ymlfile))
        self.dir = database.abspath(self.relpath)
        self.url = urljoin(database.data_dir, self.relpath, self.ymlfile)
        super().__init__(filename=self.path)

    def remove(self):
        """
        Remove library and all files.
        """
        try:
            shutil.rmtree(self.dir)
            print("library %s removed" % self.name)
        except FileNotFoundError:
            print("library %s doesn't exist" % self.name)
            raise

    def update(self):
        """
        Update the library file
        """
        database = Database()
        database.abspath(os.path.join(self.relpath, self.ymlfile), reload=True, silent=False)


class SpecLibrary(Library):

    def __init__(self, library_name):

        super().__init__(library_name)

        self.template_names = list(self.templates.keys())
        self.template_comments = list(self.templates.values())
        self.files = [t + self.file_extension for t in self.template_names]

    def download_all(self):
        """
        Download the whole library
        """

        database = Database()
        for template in self.template_names:
            database.abspath(os.path.join(self.library_name, template))

    def __repr__(self):
        description = "Spectral Library: " + self.library_name + " " + self.title
        spec_cov = "spectral coverage: " + str(self.spectral_coverage)
        units = "wave_unit: " + self.wave_unit + "  flux_unit: " + self.flux_unit
        templates = "Templates: " + str(self.template_names)

        return ' %s \n %s \n %s \n %s' % (description, spec_cov, units, templates)

    def __str__(self):
        return self.library_name


class FilterSystem(Library):
    """
        This class contains the information of a filter system

        """

    def __init__(self, filter_system):

        super().__init__(filter_system)

        self.filter_names = list(self.filters.keys())
        self.filter_comments = list(self.filters.values())
        self.files = [f + self.file_extension for f in self.filter_names]

    def download_all(self):
        """
        Download the whole library
        """

        database = Database()
        for filt in self.filter_names:
            database.abspath(os.path.join(self.filter_system, filt))

    def __repr__(self):
        description = "Filter system: " + self.filter_system + " " + self.title
        spec_cov = "spectral coverage: " + str(self.spectral_coverage)
        units = "wave_unit: " + self.wave_unit
        templates = "Templates: " + str(self.filters)

        return ' %s \n %s \n %s \n %s' % (description, spec_cov, units, templates)

    def __str__(self):
        return self.filter_system


class ExtCurvesLibrary(Library):
    """
    Class that contains the information of the a Extinction Curve Library
    """
    def __init__(self, curve_library):

        super().__init__(curve_library)

        self.curve_names = list(self.curves.keys())
        self.curve_comments = list(self.curves.values())
        self.files = [e + self.file_extension for e in self.curve_names]

    def download_all(self):
        """
        Download the whole library
        """

        database = Database()
        for curve in self.curve_names:
            database.abspath(os.path.join(self.curve_name, curve))

    def __repr__(self):
        description = "Extinction Curves: " + self.name + " " + self.title
        units = "wave_unit: " + self.wave_unit + "  flux_unit: " + self.flux_unit
        templates = "Templates: " + str(self.curves)

        return ' %s \n %s \n %s \n %s' % (description, units, templates)

    def __str__(self):
        return self.curve_name


class SpectrumContainer(SpecLibrary):
    """
    A holder of spectral template information, including template characteristics and location
    """

    def __init__(self, template):
        self.template = template
        self.template_name = os.path.basename(self.template)

        library_name = os.path.dirname(self.template)
        super().__init__(library_name=library_name)

        if self.template_name not in self.template_names:
            raise ValueError("Template %s not in library" % self.template_name)

        self.datafile = self.template_name + self.file_extension

        database = Database()

        self.path = database.abspath(os.path.join(self.relpath, self.datafile))
        self.url = urljoin(database.database_url, self.relpath, self.datafile)
        self.template_comment = self.templates[self.template_name]
        self.filename = self.path
        self._update_attrs()

    def remove(self):
        """
        remove the file
        """
        try:
            shutil.rmtree(self.path)
            print("library %s removed" % self.path)
        except FileNotFoundError:
            print("file %s doesn't exist" % self.path)
            raise

    def _update_attrs(self):
        """
        Here to just delete unnecessary stuff
        Returns
        -------
        """
        self.meta.pop("templates", None)
        self.meta.pop("summary", None)
        self.__dict__.pop("templates", None)
        self.__dict__.pop("summary", None)
        self.__dict__.pop("template_names", None)
        self.__dict__.pop("template_comments", None)
        self.__dict__.pop("ymlfile", None)

    def __repr__(self):

        s1 = "Spectral template: %s" % (self.template)

        return s1


class FilterContainer(FilterSystem):

    def __init__(self, filter_name):

        self.filter_name = filter_name

        self.basename = os.path.basename(self.filter_name)
        filter_system = os.path.dirname(self.filter_name)

        super().__init__(filter_system=filter_system)
        if self.basename not in self.filters:
            raise ValueError("Filter %s not in library" % self.filter_name)

        database = Database()

        self.datafile = self.basename + self.file_extension
        self.path = database.abspath(os.path.join(self.relpath, self.datafile))
        self.url = urljoin(database.database_url, self.relpath, self.datafile)
        self.filter_comment = self.filters[self.basename]
        self.filename = self.path

        self._update_attrs()

    def remove(self):
        """
        remove the file
        """
        try:
            shutil.rmtree(self.path)
            print("file %s removed" % self.path)
        except FileNotFoundError:
            print("file %s doesn't exist" % self.path)
            raise

    def _update_attrs(self):
        """
        Here to just delete unnecessary stuff
        """
        self.meta.pop("filterrs", None)
        self.meta.pop("summary", None)
        self.__dict__.pop("filters", None)
        self.__dict__.pop("summary", None)
        self.__dict__.pop("filter_names", None)
        self.__dict__.pop("filter_comments", None)
        self.__dict__.pop("ymlfile", None)

    def __repr__(self):
        s = "Filter: " + self.filter_name
        return s


class ExtCurveContainer(ExtCurvesLibrary):

    def __init__(self, curve_name):

        self.curve_name = curve_name
        self.basename = os.path.basename(self.curve_name)
        curve_library = os.path.dirname(self.curve_name)

        super().__init__(curve_library=curve_library)

        if self.basename not in self.curve_names:
            raise ValueError("Extinction Curve '%s' not in library" % self.curve_name)

        database = Database()

        self.datafile = self.basename + self.file_extension
        self.path = database.abspath(os.path.join(self.relpath, self.datafile))
        self.url = urljoin(database.database_url, self.relpath, self.datafile)
        self.curve_comment = self.curves[self.basename]
        self.filename = self.path

        self._update_attrs()

    def remove(self):
        """
        remove the file
        """
        try:
            shutil.rmtree(self.path)
            print("file %s removed" % self.path)
        except FileNotFoundError:
            print("file %s doesn't exist" % self.path)
            raise

    def _update_attrs(self):
        """
        Here to just delete unnecessary stuff
        """
        self.meta.pop("curves", None)
        self.meta.pop("summary", None)
        self.__dict__.pop("curves", None)
        self.__dict__.pop("summary", None)
        self.__dict__.pop("curve_names", None)
        self.__dict__.pop("curve_comments", None)
        self.__dict__.pop("ymlfile", None)

    def __repr__(self):
        s = "Extinction curve: " + self.curve_name
        return s


# TODO: functions


def get_library(library_name):
    """
    Download library and all files
    """
    pass


def get_filter_system(filter_system):
    """
    download all filters of a system
    """
    pass


def get_extcurves(extinction_curves):
    """
    Download all extinction curves
    """
    pass


def download_database():
    """
    Download
    """
    pass


def set_root_dir(path):
    """
    Set the root dir where the files will be stored.
    Probably best way is to write a small file in the data directory.


    Also get root dir should be also able to read from that file.

    Database should be get rid of initialiting parameters or use them
    to set new directories
    """
    pass


#
def database_as_table():
    """
    Show the contents of the database as table
    Returns
    -------

    """

    pass


def database_as_tree():
    """
    Show the contents od the database as a tree
    Returns
    -------

    """
    pass






# -*- coding: utf-8 -*-
import os
import inspect
from . import spextra
from . import database
from . import utils

from .spextra import Spextrum, Passband, ExtinctionCurve, get_vega_spectrum, DEFAULT_FILTERS, DEFAULT_SPECTRA, DEFAULT_CURVES
from .database import Database, SpecLibrary, FilterSystem, ExtCurvesLibrary, DefaultData
from .utils import Config

__pkg_dir__ = os.path.dirname(inspect.getfile(inspect.currentframe()))
__data_dir__ = os.path.join(__pkg_dir__, "data")




# -*- coding: utf-8 -*-
"""
speXtra: A python tool to manage and manipulate astronomical spectra
"""
import numbers
import warnings
import os

import numpy as np
from scipy.ndimage import gaussian_filter1d

import astropy.units as u
import astropy.constants as constants
from astropy.modeling.models import Scale

from synphot import (units, SourceSpectrum, SpectralElement, Observation,
                     BaseUnitlessSpectrum, ReddeningLaw, utils)
from synphot.models import (Empirical1D, GaussianFlux1D, Gaussian1D, Box1D, ConstFlux1D, BlackBody1D, PowerLawFlux1D)
from synphot.specio import read_ascii_spec, read_fits_spec
from synphot import exceptions

from .database import SpectrumContainer, FilterContainer, ExtCurveContainer, DefaultData
from .utils import download_svo_filter



__all__ = ["Spextrum", "Passband", "ExtinctionCurve",  "get_vega_spectrum"]

speed_of_light = constants.c

DEFAULT_FILTERS = DefaultData().filters
DEFAULT_SPECTRA = DefaultData().spectra
DEFAULT_CURVES = DefaultData().extcurves


class Passband(SpectralElement, FilterContainer):
    """
    This should be the holder of all information and operations related to the filters
    including path, etc.

    TODO: Implement proper __add__, __sub__, __mul__, etc

    """

    def __init__(self, filter_name=None, modelclass=None, **kwargs):

        if filter_name is not None:
            if filter_name in DEFAULT_FILTERS:
                filter_name = DEFAULT_FILTERS[filter_name]

            try:
                FilterContainer.__init__(self, filter_name)
                meta, wave, trans = self._loader()
                SpectralElement.__init__(self, Empirical1D, points=wave, lookup_table=trans, meta=meta)
            except ValueError as e1:
                try:  # try to download it from SVO
                    meta, wave, trans = self._from_svo(filter_name)
                    SpectralElement.__init__(self, Empirical1D, points=wave, lookup_table=trans,
                                             meta=meta)
                except ValueError as e2:
                    print("filter %s doesn't exist" % filter_name)

        elif modelclass is not None:
            SpectralElement.__init__(self, modelclass, **kwargs)
        else:
            raise ValueError("please define a filter")

    def _loader(self):
        """
        Load a filter from the database

        Returns
        -------
        meta: metadata of the spectra (header)
        lam: wavelengths
        flux: flux
        """

        try:  # it should also try to read it from the file directly
            self.wave_unit = units.validate_unit(self.wave_unit)
        except exceptions.SynphotError:
            self.wave_unit = u.AA

        # make try and except here to catch most problems
        if self.data_type == "fits":
            meta, lam, trans = read_fits_spec(self.filename, ext=1, wave_unit=self.wave_unit)
                                        #      wave_col=self.wave_column_name, flux_col=self.trans_column_name)
        elif self.data_type == "ascii":
            meta, lam, trans = read_ascii_spec(self.filename,  wave_unit=self.wave_unit)
                                     #          wave_col=self.wave_column_name, flux_col=self.trans_column_name)
                                             # , flux_unit=self._internal_flux_unit,
        lam = lam.to(u.AA)

        return meta, lam, trans.value

    @classmethod
    def from_vectors(cls, waves, trans, meta=None, wave_unit=u.AA):
        """
        Create a ``Passband`` directly from from vectos (lists, numpy.arrays, etc)
        Parameters
        ----------
        waves: list-like
        trans: list-like
        meta: dictionary containing the metadata
        wave_unit: u.Quantity, defaulted to angstroms

        Returns
        -------
        Passband
        """
        if isinstance(waves, u.Quantity) is False:
            waves = waves*wave_unit

        modelclass = SpectralElement(Empirical1D, points=waves, lookup_table=trans, meta=meta)

        return cls(modelclass=modelclass)

    @classmethod
    def from_file(cls, filename, **kwargs):

        modelclass = SpectralElement.from_file(filename, **kwargs)

        return cls(modelclass=modelclass)

    @classmethod
    def gaussian(cls, center, fwhm, peak, **kwargs):
        """
        Creates a filter with a gaussian shape with given user parameters
        Returns
        -------

        """
        if isinstance(center, u.Quantity) is False:
            center = center*u.AA
        if isinstance(fwhm, u.Quantity) is False:
            fwhm = fwhm*u.AA

        sigma = fwhm / (2.0 * np.sqrt(2.0 * np.log(2.0)))

        modelclass = SpectralElement(Gaussian1D, amplitude=peak, mean=center, stddev=sigma, **kwargs)

        return cls(modelclass=modelclass)

    @classmethod
    def square(cls, wmin, wmax, transmission):
        """
        Creates a filter with a rectangular shape
        Returns
        -------

        """
        if isinstance(wmin, u.Quantity) is False:
            wmin = wmin*u.AA
        if isinstance(wmax, u.Quantity) is False:
            wmax = wmax*u.AA

        center = (wmax + wmin) * 0.5
        width = wmax - wmin
        modelclass = SpectralElement(Box1D, amplitude=transmission, x_0=center, width=width)

        return cls(modelclass=modelclass)

    def _from_svo(self, filter_name):

        wave, trans = download_svo_filter(filter_name)
        meta = {"filter_name": filter_name, "source": "SVO"}
        self.filter_name = filter_name
        self.wave_unit = u.AA

        return meta, wave, trans


class ExtinctionCurve(ReddeningLaw, ExtCurveContainer):
    """
    Extinction curves

    TODO: Implement proper __add__, __sub__, __mul__, etc

    """
    def __init__(self, curve_name=None, modelclass=None, **kwargs):

        if curve_name is not None:
            if curve_name in DEFAULT_CURVES:
                curve_name = DEFAULT_CURVES[curve_name]

            ExtCurveContainer.__init__(self, curve_name)
            meta, wave, rvs = self._loader()
            ReddeningLaw.__init__(self, Empirical1D, points=wave, lookup_table=rvs, meta=meta)

        elif modelclass is not None:
            ReddeningLaw.__init__(self, modelclass)
        else:
            raise ValueError("please define a filter")

    @classmethod
    def from_vectors(cls, waves, flux, meta=None, wave_unit=u.AA):
        """
        Create a ``Passband`` directly from from vectos (lists, numpy.arrays, etc)
        Parameters
        ----------
        waves: list-like
        flux: list-like
        meta: dictionary containing the metadata
        wave_unit: u.Quantity, defaulted to angstroms
        flux_unit: u.Quantiy, defaulted to FLAM

        Returns
        -------
        Passband
        """
        if isinstance(waves, u.Quantity) is False:
            waves = waves * wave_unit

        modelclass = SpectralElement(Empirical1D, points=waves, lookup_table=flux, meta=meta)

        return cls(modelclass=modelclass)

    @classmethod
    def from_file(cls, filename, **kwargs):

        modelclass = ReddeningLaw.from_file(filename, **kwargs)

        return cls(modelclass=modelclass)

    def _loader(self):
        """
        Load a filter from the database

        Returns
        -------
        meta: metadata of the spectra (header)
        lam: wavelengths
        flux: flux
        """

        try:  # it should also try to read it from the file directly
            self.wave_unit = units.validate_unit(self.wave_unit)
        except exceptions.SynphotError:
            self.wave_unit = u.AA

        if self.data_type == "fits":
            meta, lam, rvs = read_fits_spec(self.filename, ext=1,
                                            wave_unit=self.wave_unit, flux_unit=self.extinction_unit,
                                            # self._internal_flux_unit,
                                            wave_col=self.wave_column, flux_col=self.extinction_column)
        elif self.data_type == "ascii":
            meta, lam, rvs = read_ascii_spec(self.filename,
                                             wave_unit=self.wave_unit, flux_unit=self.extinction_unit,
                                             # self._internal_flux_unit,
                                             wave_col=self.wave_column, flux_col=self.extinction_column)

        return meta, lam, rvs


class Spextrum(SpectrumContainer, SourceSpectrum):
    """
    Class to handle spectra. This class download, load, stores and manipulates the spectra.

    This class can be initialized with a remote file which will be downloaded from
    the database or with a synphot.BaseSpectrum or SourceSpectrum


    Parameters
    ----------
    template_name : Name of the template to download with format library/template e.g. "kc96/s0"
    modelclass : SourceSpectrum or BaseSpectrum

    """

    def __init__(self, template_name=None, modelclass=None, **kwargs):

        if template_name is not None:
            if template_name in DEFAULT_SPECTRA:
                template_name = DEFAULT_SPECTRA[template_name]

            SpectrumContainer.__init__(self, template_name)
            meta, lam, flux = self._loader()
            SourceSpectrum.__init__(self, Empirical1D, points=lam, lookup_table=flux, meta=meta,  **kwargs)

            self.origin = "origin: " + self.template
        elif modelclass is not None:

            SourceSpectrum.__init__(self, modelclass=modelclass, **kwargs)

        else:
            raise ValueError("please define a spectra")

    def _loader(self):
        """
        Load a template from the database

        Returns
        -------
        meta: metadata of the spectra (header)
        lam: wavelengths
        flux: flux
        """
        try:
            self.wave_unit = units.validate_unit(self.wave_unit)
            self.flux_unit = units.validate_unit(self.flux_unit)
        except exceptions.SynphotError:   # Assumes angtroms and FLAM
            self.wave_unit = u.AA
            self.flux_unit = units.FLAM

     #   self.resolution = self.resolution * self.wave_unit

        if self.data_type == "fits":  # make try and except here to catch most problems
            meta, lam, flux = read_fits_spec(self.path, ext=1,
                                             wave_unit=self.wave_unit, flux_unit=self.flux_unit,
                                             wave_col=self.wave_column_name, flux_col=self.flux_column_name)
        else:
            meta, lam, flux = read_ascii_spec(self.path,
                                              wave_unit=self.wave_unit, flux_unit=self.flux_unit)

        return meta, lam, flux

    @classmethod
    def from_arrays(cls, waves, flux, meta=None, wave_unit=u.AA, flux_unit=units.FLAM):
        """
        Create a ``Passband`` directly from from arrays (lists, numpy.arrays, etc)
        Parameters
        ----------
        waves: list-like
        flux: list-like
        meta: dictionary containing the metadata
        wave_unit: u.Quantity, defaulted to angstroms
        flux_unit: u.Quantiy, defaulted to FLAM

        Returns
        -------
        Passband
        """
        if isinstance(waves, u.Quantity) is False:
            waves = waves * wave_unit
        if isinstance(flux, (u.Quantity, u.core.Unit)) is False:
            flux = flux * flux_unit

        modelclass = SourceSpectrum(Empirical1D, points=waves, lookup_table=flux, meta=meta)

        sp = cls(modelclass=modelclass)
        sp.origin = repr(sp.model)
        return sp

    @classmethod
    def from_file(cls, filename, **kwargs):
        """
        This is just an wrapper of the ``synphot.SourceSpectrum.from_file()`` method

        Parameters
        ----------
        filename
        kwargs

        Returns
        -------
        Spextrum instance

        """
        modelclass = SourceSpectrum.from_file(filename, **kwargs)
        sp = cls(modelclass=modelclass)
        sp.origin = filename

        return sp

    @classmethod
    def from_specutils(cls, spectrum_object):
        """
        This function _tries_ to create a Spectrum from a ``specutils.Spectrum1D`` instance.

        ``specutils.Spectrum1D``  can read multiple file formats with the ``.read`` method.
        please read ``specutils`` documentation.

        Parameters
        ----------
        spectrum_object: specutils.Spectrum1D object

        Returns
        -------
        a Spextrum instance
        """
        meta = spectrum_object.meta
        lam = spectrum_object.spectral_axis
        flux = spectrum_object.flux
        modelclass = SourceSpectrum(Empirical1D, points=lam, lookup_table=flux, meta=meta)

        sp = cls(modelclass=modelclass)
        sp.origin = repr(spectrum_object)

        return sp

    @classmethod
    def flat_spectrum(cls, amplitude=0, waves=None):
        """
        Creates a flat spectrum in the preferred system scaled to a magnitude,
        default a zero magnitude spectrum
        Parameters
        ----------
        amplitude: float, u.Quantity
            amplitude/magnitude of the reference spectrum, default=0
            default is u.ABmag
            for vega use u.mag or ``synphot.units.VEGAMAG``

        waves: The waveset of the reference spectrum if not Vega
           if not provided they will be created at a resolution of R~800

        Returns
        -------
        a Spextrum instance
        """
        if waves is None:  # set a default waveset with R~805
            waves, info = utils.generate_wavelengths(minwave=100, maxwave=50000, num=5000,
                                                     log=True, wave_unit=u.AA)

        if isinstance(amplitude, u.Quantity) is False:
            amplitude = amplitude*u.ABmag

        if amplitude.unit is u.Unit("mag") or amplitude.unit is u.Unit("vegamag"):
            spec = get_vega_spectrum()
            spec = spec * 10 ** (-0.4 * amplitude.value)
            system_name = amplitude.unit
        else:
            const = ConstFlux1D(amplitude=amplitude)
            spec = cls(modelclass=Empirical1D, points=waves.value, lookup_table=const(waves.value))
            system_name = amplitude.unit

        spec.origin = "<flat spectrum, amplitude %s in system %s>" % (str(amplitude), str(system_name))

        return spec

    @classmethod
    def black_body_spectrum(cls, temperature=9500, amplitude=0, filter_curve=None, waves=None):
        """
        Produce a blackbody spectrum for a given temperature and scale it to a magnitude
        in a filter

        Parameters
        ----------
        temperature: the temperature in Kelvin degrees
        amplitude: `astropy.Quantity``, float
                The value that the spectrum should have in the given filter. Acceptable
                astropy quantities are:
                - u.mag : Vega magnitudes
                - u.ABmag : AB magnitudes
                - u.STmag : HST magnitudes
                - u.Jy : Jansky per filter bandpass
                Additionally the ``FLAM`` and ``FNU`` units from ``synphot.units`` can
                be used when passing the quantity for ``amplitude``:

        filter_curve : str
                Name of a filter from
                - a generic filter name (see ``FILTER_DEFAULTS``)
                - a spanish-vo filter service reference (e.g. ``"Paranal/HAWKI.Ks"``)
                - a filter in the spextra database
                - a filename with the filter file
                - a ``Passband`` or ``synphot.SpectralElement`` object


        Returns
        -------
        a scaled black-body spectrum
        """

        if waves is None:  # set a default waveset with R~805
            waves, info = utils.generate_wavelengths(minwave=100, maxwave=50000, num=5000,
                                                            log=True, wave_unit=u.AA)

        if isinstance(amplitude, u.Quantity) is False:
            amplitude = amplitude*u.ABmag

        bb = BlackBody1D(temperature=temperature)

        sp = cls(modelclass=Empirical1D, points=waves, lookup_table=bb(waves))
        sp = sp.scale_to_magnitude(amplitude=amplitude, filter_curve=filter_curve)
        sp.origin = "Blackbody spectrum, amplitude %s and temperature %s" % (str(amplitude), str(temperature))

        return sp

    @classmethod
    def powerlaw(cls, alpha=1, x_0=5000, amplitude=0, filter_curve=None, waves=None):
        """
        Return a power law spectrum F(lambda) ~ lambda^alpha scaled to a magnitude
        (amplitude) in an particular band

        Parameters
        ----------
        alpha : The spectral slope
        x_0 : float, u.Quantity
           Pivot wavelength
        amplitude : float, u.Quantity
            normalize the spectrum to that quantity
        filter_curve : str
                Name of a filter from
                - a generic filter name (see ``FILTER_DEFAULTS``)
                - a spanish-vo filter service reference (e.g. ``"Paranal/HAWKI.Ks"``)
                - a filter in the spextra database
                - a filename with the filter file
                - a ``Passband`` or ``synphot.SpectralElement`` object

        Returns
        -------

        """
        if waves is None:  # set a default waveset with R~805
            waves, info = utils.generate_wavelengths(minwave=100, maxwave=50000, num=5000,
                                                            log=True, wave_unit=u.AA)

        if isinstance(amplitude, u.Quantity) is False:
            amplitude = amplitude*u.ABmag

        pl = PowerLawFlux1D(amplitude, alpha=alpha, x_0=x_0, amplitude=amplitude)
        sp = cls(modelclass=Empirical1D, points=waves, lookup_table=pl(waves))

        sp = sp.scale_to_magnitude(amplitude=amplitude, filter_curve=filter_curve)

        sp.origin = "Power law spectrum: slope %s, amplitude %s in filter %s" % \
                    (str(alpha), str(amplitude), filter_curve)

        return sp

    def cut(self, wmin, wmax):
        """
        Cut the spectrum between wmin and wmax

        Parameters
        ----------
        wmin: float, u.Quantity,
        wmax: float, u.Quantity,

        Returns
        -------

        a new spextrum
        """
        if isinstance(wmin, u.Quantity) is False:
            wmin = wmin * u.AA
        if isinstance(wmax, u.Quantity) is False:
            wmax = wmax * u.AA

        wmin = wmin.to(u.AA)
        wmax = wmax.to(u.AA)
        new_waves = self.waveset[(self.waveset > wmin) & (self.waveset < wmax)]
        sp = self(new_waves)
        sp.meta.update({"wmin": wmin, "wmax": wmax})

        return sp

    def redshift(self, z=0, vel=0):
        """
        Redshift or blueshift a spectra

        Parameters
        ----------
        z: redshift or
        vel: radial velocity,  if no unit are present it is assumed to be in m/s

        Returns
        -------
        a new Spextrum instance

        """
        if vel != 0:
            if isinstance(vel, u.Quantity) is False:
                vel = vel * u.m / u.s  # assumed to be in m/s

            z = (vel.to(u.m / u.s) / speed_of_light).value
        if z <= -1:
            raise ValueError("Redshift or velocity unphysical")

        lam = self.waveset * (1 + z)
        flux = self(self.waveset)
        self.meta.update({"redshift": z})
        sp = Spextrum(modelclass=Empirical1D, points=lam, lookup_table=flux, meta=self.meta)
        sp = self._restore_attr(sp)

        return sp

    def scale_to_magnitude(self, amplitude, filter_curve=None):
        """
            Scales a Spectrum to a value in a filter
            copied from scopesim.effects.ter_curves.scale_spectrum with slight modifications
            Parameters
            ----------
            amplitude : ``astropy.Quantity``, float
                The value that the spectrum should have in the given filter. Acceptable
                astropy quantities are:
                - u.mag : Vega magnitudes
                - u.ABmag : AB magnitudes
                - u.STmag : HST magnitudes
                - u.Jy : Jansky per filter bandpass
                Additionally the ``FLAM`` and ``FNU`` units from ``synphot.units`` can
                be used when passing the quantity for ``amplitude``:

            filter_name : str
                Name of a filter from
                - a generic filter name (see ``DEFAULT_FILTERS``)
                - a spanish-vo filter service reference (e.g. ``"Paranal/HAWKI.Ks"``)
                - a filter in the spextra database
                - the path to a filename containing the curve (see ``Passband``)
                - a ``Passband`` or ``synphot.SpectralElement`` instance
            filter_file: str
                A file with a transmission curve


            Returns
            -------
            spectrum : a Spectrum
                Input spectrum scaled to the given amplitude in the given filter
        """

        if os.path.exists(filter_curve):
            filter_curve = Passband.from_file(filename=filter_curve)
        elif isinstance(filter_curve, (Passband, SpectralElement)):
            filter_curve = filter_curve
        else:
            filter_curve = Passband(filter_curve)

        if isinstance(amplitude, u.Quantity) is False:
            amplitude = amplitude*u.ABmag


       #     if amplitude.unit.physical_type == "spectral flux density":
       #         if amplitude.unit != u.ABmag:
       #             amplitude = amplitude.to(u.ABmag)
       #         ref_spec = self.flat_spectrum(amplitude=amplitude.value, system_name="AB")

        #    elif amplitude.unit.physical_type == "spectral flux density wav":
       #         if amplitude.unit != u.STmag:
       #             amplitude = amplitude.to(u.STmag)
       #         ref_spec = self.flat_spectrum(amplitude=amplitude.value, system_name="ST")

#            elif amplitude.unit == u.mag:
 #               ref_spec = self.flat_spectrum(amplitude=amplitude.value, system_name="Vega")

  #          else:
   #             raise ValueError("Units of amplitude must be one of "
    #                             "[u.mag, u.ABmag, u.STmag]: {}".format(amplitude))
    #    else:

        ref_spec = self.flat_spectrum(amplitude=amplitude)

        ref_flux = Observation(ref_spec, filter_curve).effstim(flux_unit=units.PHOTLAM)
        real_flux = Observation(self, filter_curve).effstim(flux_unit=units.PHOTLAM)

        scale_factor = ref_flux / real_flux
        sp = self * scale_factor

        return sp

    def get_magnitude(self, filter_curve=None,  system_name="AB"):
        """
            Obtain the magnitude in filter for a user specified photometric system

            Parameters
            ----------
            filter_curve : str
                Name of a filter from
                - a generic filter name (see ``DEFAULT_FILTERS``)
                - a spanish-vo filter service reference (e.g. ``"Paranal/HAWKI.Ks"``)
                - a filter in the spextra database
                - the path to the file containing the filter (see ``Passband``)
                - a ``Passband`` or ``synphot.SpectralElement`` object
            system_name: str
                The photometric system Vega, AB or ST


            Returns
            -------
            spectrum : a Spextrum
                Input spectrum scaled to the given amplitude in the given filter
        """
        if os.path.exists(filter_curve):
            filter_curve = Passband.from_file(filename=filter_curve)
        elif isinstance(filter_curve, (Passband, SpectralElement)):
            filter_curve = filter_curve
        else:
            filter_curve = Passband(filter_curve)

        if system_name.lower() in ["vega"]:
            unit = u.mag
        elif system_name.lower() in ["st", "hst"]:
            unit = u.STmag
        else:
            unit = u.ABmag

        ref_spec = self.flat_spectrum(amplitude=0*unit)
        ref_flux = Observation(ref_spec, filter_curve).effstim(flux_unit=units.PHOTLAM)
        real_flux = Observation(self, filter_curve).effstim(flux_unit=units.PHOTLAM)

        mag = -2.5*np.log10(real_flux.value/ref_flux.value)

        return mag * unit

    def photons_in_range(self, wmin=None, wmax=None, area=1*u.cm**2,
                         filter_curve=None):
        """
        Return the number of photons between wave_min and wave_max or within
        a bandpass (filter)

        Parameters
        ----------
        wmin :
            [Angstrom]
        wmax :
            [Angstrom]
        area : u.Quantity
            [cm2]
        filter_curve : str
                Name of a filter from
                - a generic filter name (see ``DEFAULT_FILTERS``)
                - a spanish-vo filter service reference (e.g. ``"Paranal/HAWKI.Ks"``)
                - a filter in the spextra database
                - the path to the file containing the filter (see ``Passband``)
                - a ``Passband`` or ``synphot.SpectralElement`` object

        Returns
        -------
        counts : u.Quantity array

        """
        if isinstance(area, u.Quantity):
            area = area.to(u.cm ** 2).value  #
        if isinstance(wmin, u.Quantity):
            wmin = wmin.to(u.Angstrom).value
        if isinstance(wmax, u.Quantity):
            wmin = wmax.to(u.Angstrom).value

        if filter_curve is None:
            # this makes a bandpass out of wmin and wmax
            try:
                mid_point = 0.5 * (wmin + wmax)
                width = abs(wmax - wmin)
                filter_curve = SpectralElement(Box1D, amplitude=1, x_0=mid_point, width=width)
            except ValueError("Please specify wmin/wmax or a filter"):
                raise

        elif os.path.exists(filter_curve):
            filter_curve = Passband.from_file(filename=filter_curve)
        elif isinstance(filter_curve, (Passband, SpectralElement)):
            filter_curve = filter_curve
        else:
            filter_curve = Passband(filter_curve)

        obs = Observation(self, filter_curve)
        counts = obs.countrate(area=area * u.cm ** 2)

        return counts

    def get_flux(self, wmin=None, wmax=None, filter_curve=None, flux_unit=units.FLAM):
        """
        Return the flux within a passband

        Parameters
        ----------
        wmin : float, u.Quantity
           minimum wavelength
        wmax : float, u.Quantity
          maximum wavelength
        filter_curve : str
                Name of a filter from
                - a generic filter name (see ``DEFAULT_FILTERS``)
                - a spanish-vo filter service reference (e.g. ``"Paranal/HAWKI.Ks"``)
                - a filter in the spextra database
                - the path to the file containing the filter (see ``Passband``)
                - a ``Passband`` or ``synphot.SpectralElement`` object
        flux_unit: synphot.units, u.Quantity

        Returns
        -------

        """

        if isinstance(wmin, u.Quantity):
            wmin = wmin.to(u.Angstrom).value
        if isinstance(wmax, u.Quantity):
            wmin = wmax.to(u.Angstrom).value

        if filter_curve is None:
            # this makes a bandpass out of wmin and wmax
            try:
                filter_curve = Passband.square(wmin=wmin, wmax=wmax)

            except ValueError("Please specify wmin/wmax or a filter"):
                raise
        elif os.path.exists(filter_curve):
            filter_curve = Passband.from_file(filename=filter_curve)
        elif isinstance(filter_curve, (Passband, SpectralElement)):
            filter_curve = filter_curve
        else:
            filter_curve = Passband(filter_curve)

        flux = Observation(self, filter_curve).effstim(flux_unit=flux_unit)

        return flux


    def add_emi_lines(self, center, fwhm, flux):
        """
        Add emission lines to an `Spextrum`

        TODO: accept different profiles (Lorentz1D, Voigt1D, etc)\

        Parameters
        ----------
        center: float, list, np.ndarray, u.Quantity
            The center of the line

        fwhm:  float, list, np.ndarray, u.Quantity
            The FWHM of the line
        flux: float, list, np.ndarray, u.Quantity
            The Equivalent Width of the line

        Returns
        -------
        Spextrum

        """
        if isinstance(center, u.Quantity) is True:
            center = center.to(u.AA).value
        if isinstance(flux, u.Quantity) is True:
            flux = flux.to(u.erg / (u.cm ** 2 * u.s)).value
        if isinstance(fwhm, u.Quantity) is True:
            fwhm = fwhm.to(u.AA).value

        centers = np.array([center]).flatten()
        fluxes = np.array([flux]).flatten()
        fwhms = np.array([fwhm]).flatten()
        sp = self
        sp.meta.update({"em_lines": {"center": list(centers),
                                     "flux": list(fluxes),
                                     "fwhm": list(fwhms)}})

        for c, f, w in zip(center, flux, fwhm):

            line = GaussianFlux1D(mean=c, total_flux=f, fwhm=w)
            lam = line.sampleset(factor_step=0.3)  # bit better than Nyquist
            g_em = SourceSpectrum(Empirical1D, points=lam, lookup_table=line(lam)) #, meta=sp.meta)

            sp = sp + g_em

        sp = self._restore_attr(Spextrum(modelclass=sp))

        return sp

    def add_abs_lines(self, center, ew, fwhm):
        """
        TODO: accept different profiles (Lorentz1D, Voigt1D, etc)
        Add a absorption line of to a spectrum with center, fwhm and equivalent width specified by the user
        It also supports emission lines if ew is negative

        Parameters
        ----------
        center: float, list, np.ndarray, u.Quantity
            The center of the line

        fwhm:  float, list, np.ndarray, u.Quantity
            The FWHM of the line
        ew: float, list, np.ndarray, u.Quantity
            The Equivalent Width of the line

        Returns
        -------
        Spextrum
        """
        if isinstance(center, u.Quantity) is True:
            center = center.to(u.AA).value
        if isinstance(ew, u.Quantity) is True:
            ew = ew.to(u.AA).value
        if isinstance(fwhm, u.Quantity) is True:
            fwhm = fwhm.to(u.AA).value

        center = np.array([center]).flatten()
        ew = np.array([ew]).flatten()
        fwhm = np.array([fwhm]).flatten()
        sp = self  #  .__class__(modelclass=self.model)

        sp.meta.update({"em_lines": {"center": list(center),
                                     "ew": list(ew),
                                     "fwhm": list(fwhm)}})

        for c, e, f in zip(center, ew, fwhm):
            sign = -1 * np.sign(e)  # to keep the convention that EL are negative and ABS are positive
            left, right = c - np.abs(e / 2), c + np.abs(e / 2)
            wavelengths = self.waveset[(self.waveset.value >= left) & (self.waveset.value <= right)]
            fluxes = units.convert_flux(wavelengths=wavelengths,
                                        fluxes=self(wavelengths),
                                        out_flux_unit=units.FLAM)
            flux = np.trapz(fluxes.value, wavelengths.value)
            line = GaussianFlux1D(mean=c, total_flux=sign * flux, fwhm=f)
            lam = line.sampleset(factor_step=0.35)  # bit better than Nyquist
            g_abs = SourceSpectrum(Empirical1D, points=lam, lookup_table=line(lam))

            sp = sp + g_abs

            if (sp(wavelengths).value < 0).any():
                warnings.warn("Warning: Flux<0 for specified EW and FHWM, setting it to Zero")
                waves = sp.waveset[sp(sp.waveset) < 0]
                zero_sp = SourceSpectrum(Empirical1D, points=waves, lookup_table=-1 * sp(waves).value)
                sp = sp + zero_sp  # Spextrum(modelclass=sp.model + zero_sp.model)

        sp = self._restore_attr(Spextrum(modelclass=sp))

        return sp

    def redden(self, curve_name,  Ebv=0, Av=None, Rv=3.1):
        """
        This function attenuate  a spectrum with a extinction curve normalized to a E(B-V)

        Parameters
        ----------
        curve_name: str, name of the extinction curve
        Av: float, Av parameter
        Ebv: float, E(B-V) color excess

        Returns
        -------
        an attenuated synphot spectrum

        """
        if Av is not None:
            Ebv = Av / Rv

        extcurve = ExtinctionCurve(curve_name)

        extinction = extcurve.extinction_curve(Ebv)
        sp = Spextrum(modelclass=self.model * extinction.model)
        sp = self._restore_attr(sp)
        sp.meta.update({"Reddening with": curve_name, "Reddening with E(B-V)": Ebv})

        return sp

    def deredden(self, curve_name, Av=None, Ebv=0, Rv=3.1):
        """
        This function de-redden a spectrum.

        Parameters
        ----------
        curve_name: str, name of the extinction curve
        Av: float, Av parameter
        Ebv: float, E(B-V) color excess

        Returns
        -------
        an attenuated synphot spectrum

        """
        if Av is not None:
            Ebv = Av / Rv

        ext_curve = ExtinctionCurve(curve_name)

        extinction = ext_curve.extinction_curve(Ebv)
        sp = Spextrum(modelclass=self.model / extinction.model)
        sp = self._restore_attr(sp)
        sp.meta.update({"Dereddening with": curve_name, "Dereddening with E(B-V)": Ebv})

        return sp

    def rebin_spectra(self, new_waves):
        """
        Rebin a synphot spectra to a new wavelength grid conserving flux.
        Grid does not need to be linear and can be at higher or lower resolution

        Parameters
        ----------
        new_waves: an array of the output wavelenghts in Angstroms but other units can be
            specified

        Returns
        -------
        a new Spextrum instance

        """
        if isinstance(new_waves, u.Quantity):
            new_waves = new_waves.to(u.AA).value

        waves = self.waveset.value   # else assumed to be angstroms
        f = np.ones(len(waves))
        filt = SpectralElement(Empirical1D, points=waves, lookup_table=f)
        obs = Observation(self, filt, binset=new_waves, force='taper')
        newflux = obs.binflux
        rebin_spec = Empirical1D(points=new_waves, lookup_table=newflux, meta=self.meta)
        sp = Spextrum(modelclass=rebin_spec)
        sp = self._restore_attr(sp)
        sp.meta.update({"rebinned_spectra": True})

        return sp

    def logrebin(self):
        """
        TODO: Estimate optimal rebinning factors here.
        Returns
        -------
        a Spextrum with a log rebinned in wavelength
        """

        waves = self.waveset.value
        steps = waves[1:] - waves[:-1]
        min_step = np.min(steps)
        wmin, wmax = np.min(waves), np.max(waves)
        print(wmin, wmax, min_step)
        nbins = int((wmax - wmin)/min_step)
        logwaves = np.geomspace(wmin, wmax, nbins)

        return self.rebin_spectra(logwaves)

    def smooth(self, sigma):
        """
        Smooth the Spectrum with a Gaussian Kernel, expressed in km/s (or other velocity unit).

        The spectra is first log-rebinned so it has a constant velocity spacing and after that
        is smoothed with a gaussian kernel


        Parameters
        ----------
        sigma: u.Quantity, width of the Kernel in

        Returns
        -------
        smoothed spextrum
        """
        if isinstance(sigma, u.Quantity) is False:
            sigma = sigma * u.km / u.s

        sigma = sigma.to(u.km / u.s)

        sp_log = self.logrebin()
        lam = sp_log.waveset.value
        flux = sp_log(sp_log.waveset)
        steps = (lam[1:] - lam[:-1]) / lam[:1]
        vel_steps = steps * speed_of_light.to(u.km/u.s)
        step_size = np.median(vel_steps)

        if step_size > sigma:
            warnings.warn("spectra is undersampled for the provided sigma value")

        conv_sigma = sigma / step_size
        smoothed_flux = gaussian_filter1d(flux, conv_sigma.value)

        self.meta.update({"KERNEL_SIZE": sigma.value})
        smoothed = Empirical1D(points=lam, lookup_table=smoothed_flux, meta=self.meta)
        sp = self._restore_attr(Spextrum(modelclass=smoothed))

        return sp


    def add_noise(self, wmin, wmax):
        """
        Returns a spectra with a given S/N in the specified wavelength range
        Returns
        -------

        """
        raise NotImplementedError

    def _restore_attr(self, spextrum):
        """
        just a hack to down stream important attributes in particular methods.
        there must be a better way
        """
        temp_dict = {k: self.__dict__[k] for k in self.__dict__ if k.startswith("_") is False}
        spextrum.__dict__.update(temp_dict)
        return spextrum


    # ------ Copied from synphot.SourceSpectrum so operations can also happen here --------

    def __add__(self, other):
        """Add ``self`` with ``other``."""
        self._validate_other_add_sub(other)
        result = self.__class__(modelclass=self.model + other.model)
        self._merge_meta(self, other, result)
        return result

    def __sub__(self, other):
        """Subtract other from self."""
        self._validate_other_add_sub(other)
        result = self.__class__(modelclass=self.model - other.model)
        self._merge_meta(self, other, result)
        return result

    def __mul__(self, other):
        """Multiply self and other."""
        self._validate_other_mul_div(other)

        if isinstance(other, (u.Quantity, numbers.Number)):
            newcls = self.__class__(modelclass=self.model | Scale(other))
        elif isinstance(other, BaseUnitlessSpectrum):
            newcls = self.__class__(modelclass=self.model * other.model)
        else:  # Source spectrum
            raise exceptions.IncompatibleSources(
                'Cannot multiply two source spectra together')

        self._merge_meta(self, other, newcls)
        return newcls

    def _validate_other_add_sub(self, other):
        """
        Conditions for other to satisfy before add/sub.
        Copied from synphot.SourceSpectrum so additions can happen with SourceSpectrum
        """
        if not isinstance(other, (self.__class__, SourceSpectrum)):
            raise exceptions.IncompatibleSources(
                'Can only operate on {0}.'.format(self.__class__.__name__))

    def _validate_other_mul_div(self, other):
        """Conditions for other to satisfy before mul/div.
           Copied from synphot.SourceSpectrum so additions can happen with a SourceSpectrum

        """

        if not isinstance(other, (u.Quantity, numbers.Number,
                                  BaseUnitlessSpectrum, SourceSpectrum, self.__class__)):
            raise exceptions.IncompatibleSources(
                'Can only operate on scalar number/Quantity or spectrum')
        elif (isinstance(other, u.Quantity) and
              (other.unit.decompose() != u.dimensionless_unscaled or
               not np.isscalar(other.value) or
               not isinstance(other.value, numbers.Real))):
            raise exceptions.IncompatibleSources(
                'Can only operate on real scalar dimensionless Quantity')
        elif (isinstance(other, numbers.Number) and
              not (np.isscalar(other) and isinstance(other, numbers.Real))):
            raise exceptions.IncompatibleSources(
                'Can only operate on real scalar number')

    def __repr__(self):

        rep = "origin " + str(self.origin)

        return rep
#------------------------------ END    -------------------------------------------


def get_vega_spectrum():
    """
    Retrieve the Vega spectrum from the database

    Notes
    -----
    To access wavelength and fluxes use::
        sp = get_vega_spectrum()
        waves, fluxes = sp.waveset, sp(sp.waveset)

    """
    return Spextrum("ref/vega")







# -*- coding: utf-8 -*-
"""
    Setup file for spextra.
"""
import sys
import setuptools
from datetime import datetime


# Version number
MAJOR = 0
MINOR = 2
ATTR = '0'

VERSION = '%d.%d%s' % (MAJOR, MINOR, ATTR)


def write_version(filename='spextra/version.py'):
    """Write a file version.py"""
    cnt = """ 
    # THIS FILE GENERATED BY speXtra setup.py
    version = '{}'
    date    = '{}' 
    """
    timestamp = datetime.utcnow().strftime('%Y-%m-%d %T GMT')
    with open(filename, 'w', encoding='utf-8') as fd:
        if sys.version_info.major == 2:
            fd.write(cnt.format(VERSION, timestamp).decode('utf-8'))
        else:
            fd.write(cnt.format(VERSION, timestamp))


with open("README.rst", "r", encoding='utf-8') as fh:
    long_description = fh.read()


def setup_sp():

    write_version()
    setuptools.setup(
        version = VERSION,
        name="speXtra",
        description="Tool to  manage and manipulate  astronomical spectra",
        long_description = long_description,
        author="Miguel Verdugo",
        author_email="mverduol@gmail.com",
        license="MIT",
        url="https://github.com/miguelverdugo/speXtra",
        package_dir={'spextra': 'spextra'},
        packages=['spextra'],
        package_data={'spextra': ['spextra/data/*']},
        include_package_data=True,
        install_requires=["numpy",
                          "astropy>=3.1",
                          "synphot>=0.2.0",
                          "pyyaml", ],
        classifiers=["Programming Language :: Python :: 3",
                     "License :: OSI Approved :: MIT License",
                     "Operating System :: OS Independent",
                     "Intended Audience :: Science/Research",
                     "Topic :: Scientific/Engineering :: Astronomy", ]
    )


if __name__ == "__main__":
    setup_sp()



import os
import inspect
import warnings
from contextlib import closing
from urllib.request import urlopen, Request
from urllib.error import URLError, HTTPError
import socket

from astropy.utils.console import ProgressBarOrSpinner
from astropy.config import get_cache_dir
import yaml


__all__ = ["Config", "download_file", "dict_generator", "download_svo_filter"]

__pkg_dir__ = os.path.dirname(inspect.getfile(inspect.currentframe()))
__data_dir__ = os.path.join(__pkg_dir__, "data")
__config_file__ = os.path.join(__data_dir__, "config.yml")


class Config:
    """
    Set up the configuration for the database.
    This class reads a yaml file in the directory data_dir that contains the default configuration of the database
    that contain default values for the following attributes

    database_url: The location of the remote database
    data_dir: The location of the database in the local hard disk, where the files are saved. Default is Null (None) which
    then lead to .astropy/cache/spextra
    remote_timeout: The time before timeout

    Normally, the user will not need to call this class, except when chaning

    If the user needs to change these values the class can be simply called as

    >> Config(data_dir="path_to_new_data_directory")

    and the new path will be used from that moment on
    """

    def __init__(self, data_dir=None, database_url=None, remote_timeout=None):

        self.config_file = __config_file__

        with open(self.config_file) as f:
            self.meta = yaml.safe_load(f)
            self.data_dir = self.meta["data_dir"]
            self.database_url = self.meta["database_url"]
            self.remote_timeout = self.meta["remote_timeout"]
            self.default_data_dir = os.path.join(get_cache_dir(), self.meta["default_data_dir"])

        if data_dir is not None:
            self._set_param("data_dir", data_dir)

        if database_url is not None:
            self._set_param("database_url", database_url)

        if remote_timeout is not None:
            self._set_param("remote_timeout", remote_timeout)

    def get_data_dir(self):
        """
        Retrieve the current data directory where the data from the database is stored
        Check whether the directory exists and if not it will create it. Raises a FileExistsError or PermissionError
        if not possible

        Returns
        -------
        path: str
            Path to the local data directory

        Raises
        ------
        FileExistsError
             if it is not a directory
        PermissionError
             if not possible to create the directory
        """
        if self.data_dir is None:
            self.data_dir = self.default_data_dir

        if os.path.isdir(self.data_dir) is False:
            try:
                os.mkdir(self.data_dir)
            except FileExistsError as e:
                raise("{0} not a directory".format(self.data_dir), e)
            except PermissionError as e :
                raise("Cannot create {0}".format(self.data_dir), e)

        return self.data_dir

    def get_database_url(self):
        """
        Retrieve the database url and check whether is reachable

        Returns
        -------
        url: str
           the url location of the database
        """
        loc = self.database_url
        try:
            request = Request(loc)
            request.get_method = lambda: 'HEAD'
            urlopen(request)
            is_url = True
        except URLError:
            is_url = False
        except ValueError:
            is_url = False

        if is_url is False:
            warnings.warn("Database might not be reachable")

        return loc

    def _set_param(self, name, value):
        d = {name: value}
        self.meta.update(d)
        self.__dict__.update(d)
        with open(self.config_file, "w") as f:
            yaml.dump(self.meta, f, sort_keys=False)

    def __repr__(self):
        return yaml.dump(self.meta, indent=4, sort_keys=False, default_flow_style=False)


def _download_file(remote_url, target, silent=False):
    """
    Accepts a URL, downloads the file to a given open file object.
    This is a modified version of astropy.utils.data.download_file that
    downloads to an open file object instead of a cache directory.
    """

    conf = Config()
    timeout = conf.remote_timeout
    download_block_size = 32768

    msg_file = None
    if silent is True:  # write the messages to dev/null
        msg_file = open(os.devnull, "w")

    try:
        # Pretend to be a web browser (IE 6.0). Some servers that we download
        # from forbid access from programs.
        headers = {'User-Agent': 'Mozilla/5.0',
                   'Accept': ('text/html,application/xhtml+xml,'
                              'application/xml;q=0.9,*/*;q=0.8')}
        req = Request(remote_url, headers=headers)
        with closing(urlopen(req, timeout=timeout)) as remote:

            # get size of remote if available (for use in progress bar)
            info = remote.info()
            size = None
            if 'Content-Length' in info:
                try:
                    size = int(info['Content-Length'])
                except ValueError:
                    pass

            dlmsg = "Downloading {0}".format(remote_url)

            with ProgressBarOrSpinner(size, dlmsg, file=msg_file) as p:
                bytes_read = 0
                block = remote.read(download_block_size)
                while block:
                    target.write(block)
                    bytes_read += len(block)
                    p.update(bytes_read)
                    block = remote.read(download_block_size)

    # Append a more informative error message to HTTPErrors, URLErrors.
    except HTTPError as e:
        e.msg = "{}. requested URL: {!r}".format(e.msg, remote_url)
        raise
    except URLError as e:
        append_msg = (hasattr(e, 'reason') and hasattr(e.reason, 'errno') and
                      e.reason.errno == 8)
        if append_msg:
            msg = "{0}. requested URL: {1}".format(e.reason.strerror,
                                                   remote_url)
            e.reason.strerror = msg
            e.reason.args = (e.reason.errno, msg)
        raise e

    # This isn't supposed to happen, but occasionally a socket.timeout gets
    # through.  It's supposed to be caught in `urrlib2` and raised in this
    # way, but for some reason in mysterious circumstances it doesn't. So
    # we'll just re-raise it here instead.
    except socket.timeout as e:
        # add the requested URL to the message (normally just 'timed out')
        e.args = ('requested URL {!r} timed out'.format(remote_url),)
        raise URLError(e)


def download_file(remote_url, local_name, silent=False):
    """
    Download a remote file to local path
    ----------
    remote_url : str
        The URL of the file to download
    local_name : str
        Absolute path filename of target file.
    Raises
    ------
    URLError
        Whenever there's a problem getting the remote file.
    """
    # ensure target directory exists
    dn = os.path.dirname(local_name)
    if os.path.exists(dn) is False:
        os.makedirs(dn)

    try:
        with open(local_name, 'wb') as target:
            _download_file(remote_url, target, silent=silent)
    except:  # noqa
        # in case of error downloading, remove file.
        if os.path.exists(local_name):
            os.remove(local_name)
            raise


def download_svo_filter(filter_name):
    """
    Query the SVO service for the true transmittance for a given filter
    Parameters
    ----------
    filter_name : str
        Name of the filter as available on the spanish VO filter service
        e.g: ``Paranal/HAWKI.Ks``

    Returns
    -------
    filt_curve : tuple with wave and trans values
    """
    from astropy.table import Table

    conf = Config()
    data_dir = conf.get_data_dir()

    origin = 'http://svo2.cab.inta-csic.es/'\
             'theory/fps3/fps.php?ID={}'.format(filter_name)

    local_path = os.path.join(data_dir, "svo_filters", filter_name)

    if os.path.exists(local_path) is False:
        download_file(origin, local_path)

    tbl = Table.read(local_path, format='votable')  # raises ValueError if table is malformed
                                                    # this can be used to catch problmes
    wave = tbl['Wavelength']
    trans = tbl['Transmission']

    return wave, trans


def dict_generator(indict, pre=None):
    """
    Make a generator out of a dictionary
    Parameters
    ----------
    indict: dict
    pre

    Yields
    ------
    list
    """

    pre = pre[:] if pre else []
    if isinstance(indict, dict):
        for key, value in indict.items():
            if isinstance(value, dict):
                for d in dict_generator(value, pre + [key]):
                    yield d
            elif isinstance(value, list) or isinstance(value, tuple):
                for v in value:
                    for d in dict_generator(v, pre + [key]):
                        yield d
            else:
                yield pre + [key]
    else:
        yield pre + [indict]


# ------

def get_filter_systems():
    """
    Return a set of the different filter system available

    Returns
    -------

    """
    try:
        import tynt
    except ImportError as e:
        print(e, "this function requires tynt. ")
    filters = tynt.FilterGenerator().available_filters()
    systems = {f.split("/")[0] for f in filters}
    return systems


def get_filter_names(system=None):
    """
    This function just returns the filters available from tynt
    if system= None returns all

    Returns
    -------

    """
    try:
        import tynt
    except ImportError as e:
        print(e, "this function requires tynt. ")
    filter_list = tynt.FilterGenerator().available_filters()
    ord_list = [[f for f in filter_list if s in f] for s in get_filter_systems()]
    flat_list = [item for sublist in ord_list for item in sublist]

    if system is not None:
        flat_list = [f for f in filter_list if system in f]

    return flat_list
